@@ -7,6 +7,7 @@
 #include <sys/mman.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <termios.h>
 #include <errno.h>
 #include <signal.h>
 #include <assert.h>
@@ -24,12 +25,6 @@ struct device_infos {
 static int memory_unsafe = 0;
 static char *default_backend = NULL;
 
-static void catch_alarm(int sig_num)
-{
-       fprintf(stderr, "Operation timed out. Exiting.\n");
-       exit(0);
-}
-
 static int setup_enter(struct setup_backend *backend)
 {
 	int r;
@@ -72,6 +67,75 @@ static int setup_leave(struct setup_backend *backend)
 	return 0;
 }
 
+static int untimed_read(int fd, char *pass, size_t maxlen)
+{
+	ssize_t i;
+
+	i = read(fd, pass, maxlen);
+	if (i > 0) {
+		pass[i-1] = '\0';
+		i = 0;
+	}
+	return i;
+}
+
+static int timed_read(int fd, char *pass, size_t maxlen, long timeout)
+{
+	struct timeval t;
+	fd_set fds;
+	int failed = -1;
+
+	FD_ZERO(&fds);
+	FD_SET(fd, &fds);
+	t.tv_sec = timeout;
+	t.tv_usec = 0;
+
+	if (select(fd+1, &fds, NULL, NULL, &t) > 0)
+		failed = untimed_read(fd, pass, maxlen);
+	else
+		fprintf(stderr, "Operation timed out.\n");
+	return failed;
+}
+
+static int interactive_pass(const char *prompt, char *pass, size_t maxlen,
+		long timeout)
+{
+	struct termios orig, tmp;
+	int failed = -1;
+	int infd, outfd;
+
+	if (maxlen < 1)
+		goto out_err;
+
+	/* Read and write to /dev/tty if available */
+	if ((infd = outfd = open("/dev/tty", O_RDWR)) == -1) {
+		infd = STDIN_FILENO;
+		outfd = STDERR_FILENO;
+	}
+
+	if (tcgetattr(infd, &orig)) {
+		set_error("Unable to get terminal");
+		goto out_err;
+	}
+	memcpy(&tmp, &orig, sizeof(tmp));
+	tmp.c_lflag &= ~ECHO;
+
+	write(outfd, prompt, strlen(prompt));
+	tcsetattr(infd, TCSAFLUSH, &tmp);
+	if (timeout)
+		failed = timed_read(infd, pass, maxlen, timeout);
+	else
+		failed = untimed_read(infd, pass, maxlen);
+	tcsetattr(infd, TCSAFLUSH, &orig);
+
+out_err:
+	if (!failed)
+		write(outfd, "\n", 1);
+	if (infd != STDIN_FILENO)
+		close(infd);
+	return failed;
+}
+
 /*
  * Password reading behaviour matrix of get_key
  * 
@@ -82,30 +146,39 @@ static int setup_leave(struct setup_backend *backend)
  * from binary file | N | N | N | Inf or options->key_size
  *
  * Legend: p..prompt, v..can verify, n..newline-stop, h..read horizon
- * 
+ *
+ * Note: --key-file=- is interpreted as a read from a binary file (stdin)
+ *
+ * Returns true when more keys are available (that is when password
+ * reading can be retried as for interactive terminals).
  */
 
-static void get_key(struct crypt_options *options, char *prompt, char **key, int *passLen) 
+static int get_key(struct crypt_options *options, char *prompt, char **key, int *passLen)
 {
 	int fd;
 	const int verify = options->flags & CRYPT_FLAG_VERIFY;
 	const int verify_if_possible = options->flags & CRYPT_FLAG_VERIFY_IF_POSSIBLE;
 	char *pass = NULL;
 	int newline_stop;
-	int read_horizon; 
+	int read_horizon;
 	
-	if(options->key_file) {
+	if(options->key_file && !strcmp(options->key_file, "-")) {
+		/* Allow binary reading from stdin */
+		fd = options->passphrase_fd;
+		newline_stop = 0;
+		read_horizon = 0;
+	} else if (options->key_file) {
 		fd = open(options->key_file, O_RDONLY);
 		if (fd < 0) {
 			char buf[128];
 			set_error("Error opening key file: %s",
-			          strerror_r(errno, buf, 128));
+				  strerror_r(errno, buf, 128));
 			goto out_err;
-		}	
+		}
 		newline_stop = 0;
 
-		/* This can either be 0 (LUKS) or the actually number 
-		   of key bytes (default or passed by -s) */
+		/* This can either be 0 (LUKS) or the actually number
+		 * of key bytes (default or passed by -s) */
 		read_horizon = options->key_size;
 	} else {
 		fd = options->passphrase_fd;
@@ -113,32 +186,23 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 		read_horizon = 0;   /* Infinite, if read from terminal or fd */
 	}	
 
-	signal(SIGALRM, catch_alarm);
-	if(options->timeout) {
-		alarm(options->timeout);
-	} else {
-		alarm(0);
-	}
-	
 	/* Interactive case */
 	if(isatty(fd)) {
-		char *pass2;
-		
-		pass2 = getpass(prompt);
-		if (!pass2) {
+		int i;
+
+		pass = safe_alloc(512);
+		if (!pass || (i = interactive_pass(prompt, pass, 512, options->timeout))) {
 			set_error("Error reading passphrase");
 			goto out_err;
 		}
-		pass = safe_strdup(pass2);
-		memset(pass2, 0, strlen(pass2));
-		
 		if (verify || verify_if_possible) {
-			char *pass_verify = getpass("Verify passphrase: ");
-			if (!pass_verify || strcmp(pass, pass_verify) != 0) {
+			char pass_verify[512];
+			i = interactive_pass("Verify passphrase: ", pass_verify, sizeof(pass_verify), options->timeout);
+			if (i || strcmp(pass, pass_verify) != 0) {
 				set_error("Passphrases do not match");
 				goto out_err;
 			}
-			memset(pass_verify, 0, strlen(pass_verify));
+			memset(pass_verify, 0, sizeof(pass_verify));
 		}
 		*passLen = strlen(pass);
 		*key = pass;
@@ -160,7 +224,7 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 		 * such as /dev/random, because in this case, the loop
 		 * will read forever.
 		 */ 
-		if(options->key_file && read_horizon == 0) {
+		if(options->key_file && strcmp(options->key_file, "-") && read_horizon == 0) {
 			struct stat st;
 			if(stat(options->key_file, &st) < 0) {
 		 		set_error("Can't stat key file");
@@ -191,17 +255,16 @@ static void get_key(struct crypt_options *options, char *prompt, char **key, int
 		pass[i] = 0;
 		*key = pass;
 		*passLen = i;
-
-		alarm(0);
 	}
-	return;
+
+	return isatty(fd); /* Return true, when password reading can be tried on interactive fds */
 
 out_err:
 	if(pass)
 		safe_free(pass);
 	*key = NULL;
 	*passLen = 0;
-	return;
+	return 0;
 }
 
 /*
@@ -215,7 +278,7 @@ static char *process_key(struct crypt_options *options, char *pass, int passLen)
 	char *key = safe_alloc(options->key_size);
 
 	/* key is coming from binary file */
-	if (options->key_file) {
+	if (options->key_file && strcmp(options->key_file, "-")) {
 		if(passLen < options->key_size) {
 			set_error("Could not read %d bytes from key file",
 			          options->key_size);
@@ -225,7 +288,7 @@ static char *process_key(struct crypt_options *options, char *pass, int passLen)
 		return key;
 	}
 	
-	/* key is coming from tty or fd */
+	/* key is coming from tty, fd or binary stdin */
 	if (options->hash) {
 		if (hash(NULL, options->hash, key,
 		         options->key_size, pass) < 0)
@@ -301,7 +364,7 @@ out:
 }
 
 static int parse_into_name_and_mode(const char *nameAndMode, char *name,
-					char *mode)
+				    char *mode)
 {
 	// Token content stringification, see info cpp/stringification
 #define str(s) #s
@@ -313,7 +376,7 @@ static int parse_into_name_and_mode(const char *nameAndMode, char *name,
 
 	if(sscanf(nameAndMode,scanpattern1, name, mode) != 2) {
 		if((r = sscanf(nameAndMode,scanpattern2,name)) == 1) {
-			strncpy(mode,"cbc-essiv:sha256",10);
+			strncpy(mode,"cbc-plain",10);
 		} 
 		else {
 			fprintf(stderr, "no known cipher-spec pattern detected\n");
@@ -387,7 +450,14 @@ static int __crypt_create_device(int reload, struct setup_backend *backend,
 	safe_free(key);
 	
 	if (!processed_key) {
-		set_error("Key processing error");
+		const char *error=get_error();
+		if(error) {
+			char *c_error_handling_sucks;
+			asprintf(&c_error_handling_sucks,"Key processing error: %s",error);
+			set_error(c_error_handling_sucks);
+			free(c_error_handling_sucks);
+		} else
+			set_error("Key processing error");
 		return -ENOENT;
 	}
 	
@@ -504,7 +574,7 @@ static int __crypt_luks_format(int arg, struct setup_backend *backend, struct cr
 	r = parse_into_name_and_mode(options->cipher, cipherName, cipherMode);
 	if(r < 0) return r;
 
-	r = LUKS_generate_phdr(&header,&mk,cipherName, cipherMode,LUKS_STRIPES);
+	r = LUKS_generate_phdr(&header,&mk,cipherName, cipherMode,LUKS_STRIPES, options->align_payload);
 	if(r < 0) { 
 		set_error("Can't write phdr");
 		return r; 
@@ -522,7 +592,6 @@ static int __crypt_luks_format(int arg, struct setup_backend *backend, struct cr
 	if(!password) {
 		r = -EINVAL; goto out;
 	}
-	
 	r = LUKS_set_key(options->device, 0, password, passwordLen, &header, &mk, backend);
 	if(r < 0) goto out; 
 
@@ -539,9 +608,18 @@ static int __crypt_luks_open(int arg, struct setup_backend *backend, struct cryp
 	struct luks_phdr hdr;
 	char *password; int passwordLen;
 	struct device_infos infos;
+	struct crypt_options tmp = {
+		.name = options->name,
+	};
 	char *dmCipherSpec;
-	int r;
+	int r, tries = options->tries;
 	
+	r = backend->status(0, &tmp, NULL);
+	if (r >= 0) {
+		set_error("Device already exists");
+		return -EEXIST;
+	}
+
 	if (get_device_infos(options->device, &infos) < 0) {
 		set_error("Can't get device information.\n");
 		r = -ENOTBLK; goto out;
@@ -549,8 +627,14 @@ static int __crypt_luks_open(int arg, struct setup_backend *backend, struct cryp
 	if (infos.readonly)
 		options->flags |= CRYPT_FLAG_READONLY;
 
+start:
 	options->key_size = 0; // FIXME, define a clean interface some day.
-	get_key(options,"Enter LUKS passphrase: ",&password,&passwordLen);
+
+	if(get_key(options,"Enter LUKS passphrase: ",&password,&passwordLen))
+		tries--;
+	else
+		tries = 0;
+
 	if(!password) {
 		r = -EINVAL; goto out;
 	}
@@ -584,9 +668,13 @@ static int __crypt_luks_open(int arg, struct setup_backend *backend, struct cryp
  out2:
 	free(dmCipherSpec);
  out1:
-	safe_free(password);	
+	safe_free(password);
  out:
 	memset(&mk,0,sizeof(mk));
+
+	if (r != 0 && tries > 0)
+		goto start;
+
 	return r;
 }
 
@@ -599,7 +687,7 @@ static int __crypt_luks_add_key(int arg, struct setup_backend *backend, struct c
 	const char *device = options->device;
 	struct crypt_options optionsCheck = { 
 		.key_file = options->key_file,
-		.flags = options->flags & ~CRYPT_FLAG_VERIFY,
+		.flags = options->flags & ~(CRYPT_FLAG_VERIFY | CRYPT_FLAG_VERIFY_IF_POSSIBLE),
 	};
 	struct crypt_options optionsSet = { 
 		.key_file = options->new_key_file,
@@ -648,6 +736,40 @@ out:
 	return r;
 }
 
+static int __crypt_luks_del_key(int arg, struct setup_backend *backend, struct crypt_options *options)
+{
+	struct luks_masterkey mk;
+	struct luks_phdr hdr;
+	char *password=NULL; 
+	unsigned int passwordLen;
+	const char *device = options->device;
+	int keyIndex = options->key_slot;
+	int openedIndex;
+	int r;
+	
+	if(options->flags & CRYPT_FLAG_VERIFY_ON_DELKEY) {
+		options->flags &= ~CRYPT_FLAG_VERIFY_ON_DELKEY;
+		options->key_size = 0; // FIXME, define a clean interface some day.
+		get_key(options,"Enter any remaining LUKS passphrase: ",&password,&passwordLen);
+		if(!password) {
+			r = -EINVAL; goto out;
+		}
+		openedIndex = LUKS_open_any_key(device, password, passwordLen, &hdr, &mk, backend);
+		if(openedIndex < 0 || keyIndex == openedIndex) {
+			printf("No remaining key available with this passphrase.\n");
+			r = -EPERM; goto out;
+		}
+	}
+	r = LUKS_del_key(device, keyIndex);
+	if(r < 0) goto out;
+
+	r = 0;
+out:
+	safe_free(password);
+	return r;
+}
+
+
 static int crypt_job(int (*job)(int arg, struct setup_backend *backend,
                                 struct crypt_options *options),
                      int arg, struct crypt_options *options)
@@ -715,7 +837,7 @@ int crypt_luksOpen(struct crypt_options *options)
 
 int crypt_luksDelKey(struct crypt_options *options)
 {
-	return LUKS_del_key(options->device, options->key_slot);
+	return crypt_job(__crypt_luks_del_key, 0, options);
 }
 
 int crypt_luksAddKey(struct crypt_options *options)
